package org.SplitPeaks;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;

public abstract class Parser {
	protected BufferedReader m_bufRead;
	protected String m_wigType;
	protected long m_currentOffset;
	
	public Parser(BufferedReader reader,String type) throws IOException 
	{
		m_bufRead = reader;
    	m_bufRead.mark(1000);
    	m_wigType = "";
    	m_currentOffset=0;
    	
        String line = m_bufRead.readLine();
        while(line!=null && !line.equals("")){
        	if(type.equals("WIG"))
        	{
        		if(line.contains("variableStep")){
            		m_bufRead.reset();
            		break;
            	}
        		else if(line.contains("fixedStep"))
        		{
        			throw new IOException ("Wig fixedStep format is not supported\n");
        		}
        		else if(line.contains("=") || line.contains("#"))  
        		{
        			m_bufRead.mark(1000);
        			line = m_bufRead.readLine();
        			continue;
        		}
        		else
        		{
        			throw new IOException ("The variableStep definition is missing in wig file\n");
        		}
        	}
        	else //BED format
        	{
        		if(line.contains("=") || line.contains("#"))  
        		{
        			m_bufRead.mark(1000);
        			line = m_bufRead.readLine();
        			continue;
        		}
        		else
        		{
        			String[] myelements = split(line);
        			if(myelements.length==4)
        			{
        				try {
                			Integer.parseInt(myelements[1]);
                			Integer.parseInt(myelements[2]);
                			}
        				catch(NumberFormatException e){
        					throw new IOException ("Invalid bedGraph file format in line starting with \"" + line + "\"" + "\n");
        				}
        			}
        			else
        			{
        				throw new IOException ("Invalid bedGraph file format in line starting with \"" + line + "\"" + "\n");
        			}
        			m_bufRead.reset();
        			break;
        		}
        	}
        }
	}

	public abstract Position nextPosition() throws IOException; 
	public abstract Position nextChrom(String curChr) throws IOException;
	
	public void close () throws FileNotFoundException,IOException {
		m_bufRead.close();
        }
	
	 public void Lineback() throws IOException
	 {
		 m_bufRead.reset();
	 }
	 
	 protected String[] split(String line)
		{
			String[] temptokens = line.split("[\t\\s\"]");
			ArrayList<String> tokens = new ArrayList<String>();
			//remove empty entries
			for(int i = 0; i < temptokens.length; ++i) 
			{
				  if(temptokens[i] != null && !temptokens[i].equals("")) 
				  {
					  tokens.add(temptokens[i]);
				  }
			}
			String[] copytokens = new String[tokens.size()];
			return (String[]) (tokens.toArray(copytokens));
		}
	 
	 public static String getFileFormat(BufferedReader reader) throws IOException
	 {
		 String wigType = "";
		// find the first line of data or variable step chromosome definition
    	for(int i =0; i<20;i++)
    	{
    		String line = reader.readLine();
    		if(line!=null && !line.equals(""))
    		{
    			String[] myelements = line.trim().split("[\\s\t]");
        		for(String s : myelements)
        		{
        			if(s.toLowerCase().contains("type"))
        			{
        				wigType = s.split("=")[1];
        				return wigType;
        			}
        		}
    		}
    		
    	}
    	return wigType;
		    	
	 }
	 public long getCurrentOffset()
	 {
		return m_currentOffset;
	 }
}
