package org.SplitPeaks;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.HashMap;


public class PeakParser {

	private RandomAccessFile m_bufRead;
	private HashMap<String, Long> m_chrInFile;
	private Position current;
	private long m_currentOffset;

	public PeakParser(RandomAccessFile file) throws FileNotFoundException,IOException {

            m_bufRead = file;
            m_chrInFile = new HashMap<String, Long>();
            long firstPeak = file.getFilePointer();
            long offset = 0;
            // Read first line
            String line = m_bufRead.readLine();
            //check if first line starts with the word track
            if(line.contains("track")){
            	firstPeak = m_bufRead.getFilePointer();
            	line = m_bufRead.readLine();
            }
            String[] myelements = this.split(line);
            if(myelements.length < 3) 
            {
            	throw new IOException ("Invalid peak file format");
            }
            else{
	            	while(line!=null && !line.equals(""))
	            	{
	            		myelements = this.split(line);
	            		String chr = myelements[0];
	            		if(!chr.toLowerCase().startsWith("chr"))
	            		{
	            			chr = "chr"+chr;
	            		}
	            		else if(chr.startsWith("CHR"))
	            		{
	            			chr="chr"+chr.substring(3);
	            		}
	            		
	            		if(!m_chrInFile.containsKey(chr))
	            		{
	            			m_chrInFile.put(chr, offset);
	            		}
	            		offset = m_bufRead.getFilePointer();
	            		line=m_bufRead.readLine();
	            	}
            	}
            
            //go to the start of the file
            m_bufRead.seek(firstPeak);
            current = null;
		}
	
	public Position nextPosition() throws Exception {
		String line = m_bufRead.readLine();
		if(line==null || line.equals(""))
		{
			return null;
		}
		m_currentOffset += line.length();
		String[] myelements = this.split(line);
		String chr="";
		int start=-1;
		int end=-1;
		try
		{
			 chr = myelements[0];
			 start = Integer.parseInt(myelements[1]);
			 end = Integer.parseInt(myelements[2]);	
		}
		catch (Exception e)
		{
			throw new IOException("Invalid peak file format in line starting with \"" + line + "\"" + "\n");
		}
		Position g = new Position(start,end,chr);
        current = g;
        return g;
		
	}
	public Position getCur() {
		return current;
	}
	public void moveToChr(String chr) throws ItemNotFoundException,IOException {
		if(m_chrInFile.containsKey(chr))
		{
			long loc = m_chrInFile.get(chr);
			m_bufRead.seek(loc);
		}
		else
		{
			throw new ItemNotFoundException("No peaks were found in chromosome " + chr);
		}
	}
	protected String[] split(String line)
	{
		String[] temptokens = line.split("[\t\\s\"]");
		ArrayList<String> tokens = new ArrayList<String>();
		//remove empty entries
		for(int i = 0; i < temptokens.length; ++i) 
		{
			  if(temptokens[i] != null && !temptokens[i].equals("")) 
			  {
				  tokens.add(temptokens[i]);
			  }
		}
		String[] copytokens = new String[tokens.size()];
		return (String[]) (tokens.toArray(copytokens));
	}
	
	public long getCurrentOffset()
	{
		return m_currentOffset;
	}	
}
