package org.SplitPeaks;


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.io.File;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.Infra.*;


public class Split {
	private float m_bd;
	private FileWriter m_fw;
	private ArrayList<SubPeak> m_SortedList = new ArrayList<SubPeak>();
	private String m_OutDir;
	private String m_PeakFileName;
	private String m_PeakFileNoExtension;
	private String m_PeakFileExtension;
	private int m_numOfseq;
	private int m_cutoff;
	private String m_SubpeakFileName;
	private String m_prefix;
	
	 // Create the listener list
    protected javax.swing.event.EventListenerList unzipListenerList =
        new javax.swing.event.EventListenerList();
    protected javax.swing.event.EventListenerList readListenerList =
        new javax.swing.event.EventListenerList();

	public Split(){}
    
    public void init (String peakFile, String wigFile,float bd, int cutoff,String outDir,int numSeq,String prefix) throws Exception {
    	if(!prefix.equals(""))
		{
			m_prefix = prefix+".";
		}
		else {m_prefix=prefix;}
    	
    	m_numOfseq = numSeq;
		m_cutoff=cutoff;
		
		File dir = new File(outDir);
		if(!dir.exists())
		{
			throw new IOException("error occurred: " + outDir + " does not exist\n");
		}
		if(!dir.isDirectory())
		{
			throw new IOException("error occurred: "+ outDir + " is not a directory\n");
		}
		File temp = new File(peakFile);
		if(!temp.exists())
		{
			throw new IOException("error occurred: "+ peakFile + " does not exist\n");
		}
		temp = new File(wigFile);
        if(!temp.exists())
		{
			throw new IOException("error occurred: " +wigFile + " does not exist\n");
		}
		
		if(outDir.endsWith(File.separator))
		{
			m_OutDir = outDir;
		}
		else
		{
			m_OutDir = outDir + File.separatorChar;
		}
		
		File pFile = new File(peakFile);
		File wig = new File(wigFile);
		
		//keep output file names
        //find extension
		//check if peakFile name contain the full path, if so separate between the path and the name
		String ParentPath = pFile.getParent();
		if(ParentPath==null)
		{
			m_PeakFileName = peakFile;
		}
		else
		{
			m_PeakFileName = pFile.getName();
		}
		
		int pos = m_PeakFileName.lastIndexOf('.'); 
        if(pos > 0)
        {
        	m_PeakFileNoExtension = m_PeakFileName.substring(0, pos);
        	m_PeakFileExtension = m_PeakFileName.substring(pos);
        }
        else
        {
        	m_PeakFileNoExtension = m_PeakFileName;
        	m_PeakFileExtension = "";
        }
        m_SubpeakFileName = m_OutDir + m_prefix + m_PeakFileNoExtension+".subpeaks"+m_PeakFileExtension;
		
        m_fw = new FileWriter(m_SubpeakFileName);
        m_fw.write("Chromosome\tStart\tEnd\tHeight\tSummitPosition\n");
        
        if(wig.isDirectory()){
			splitPeaksWigDirectory(peakFile,wigFile,bd);
		}
		else {splitPeaks(peakFile,wigFile,bd);}
		
	}
	public void splitPeaks(String peakFile, String wigFile,float bd) throws Exception {
		m_bd = bd;
		boolean isZip=false;
		
		File peakfile = new File(peakFile);
		if(!peakfile.exists()){
			throw new FileNotFoundException();
		}
		RandomAccessFile buf = new RandomAccessFile(peakfile,"rw");
		PeakParser peak = new PeakParser(buf);
	
		 if(wigFile.endsWith(".gz") || wigFile.endsWith("zip")){
			 System.out.println("Unzipping " + wigFile); 
			   String unzipFileName = this.Unzip(wigFile,true);
			   wigFile = unzipFileName;
			   isZip=true;
		   }
		
		File curwigfile = new File(wigFile);
		if(!curwigfile.exists()){
			throw new FileNotFoundException();
		}
		
		long totalLength = curwigfile.length();
		int times=0;
		fireReadProgressEvent(0);
		
		FileReader wigfile = new FileReader(wigFile);
		BufferedReader wigbuf = new BufferedReader(wigfile);
		String wigType = Parser.getFileFormat(wigbuf);
		Parser wig = null;
		if(wigType.equals("wiggle_0"))
		{
			wig = new VariableStepParser(wigbuf);
		}
		else if(wigType.equals("bedGraph"))
		{
			 wig = new BedgraphParser(wigbuf);
		}
		else
		{
			if(wigType.equals(""))
			{
				throw new IOException ("error occurred: can't guess the wig file's type, track type definition is missing\n");
			}
			else
			{
				throw new IOException("error occurred: unknown wig file type" + wigType+"\n");
			}
		}
		
		Position curwig= wig.nextPosition();
		Position curPeak = null;
		String curChr = null;
		System.out.println("Processing...");
		while(curwig!=null)
		{
			times++;
			int progress = (int)(wig.getCurrentOffset() *100 / totalLength);
        	fireReadProgressEvent(progress);
        	
			if(!curwig.getChr().equals(curChr)){
				try{
					peak.moveToChr(curwig.getChr());
					System.out.println("Start processing " + curwig.getChr() + " peaks");
				}
				catch (ItemNotFoundException e){
					curwig = nextChrom(wig,peak,curwig);
				}
				if(curwig==null)
				{
					break;
				}
				curChr = curwig.getChr();
				curPeak = peak.nextPosition();
			}
			while(curPeak!=null && curwig!=null)
			{
				if(!curPeak.getChr().equals(curChr)){
					System.out.println("End processing " + curChr + " peaks");
					curwig=nextChrom(wig, peak, curwig);
					if(curwig==null)
					{
						break;
					}
					curChr=curwig.getChr();
					curPeak = peak.nextPosition();
				}
				splitPeak(curPeak,wig);
				curPeak = peak.nextPosition();
			}
			if(curPeak==null)
			{
				System.out.println("End processing " + curChr + " peaks");
				curwig=nextChrom(wig, peak, curwig);
				if(curwig==null)
				{
					break;
				}
				curChr=curwig.getChr();
				curPeak = peak.nextPosition();
			}
		}
		wigbuf.close();
		if(isZip){
			curwigfile.delete();
		}
		fireReadProgressEvent(100);	
	}
	
	public Position nextChrom(Parser wig, PeakParser peak,Position curwig) throws Exception
	{
		//loop until we find a chromosome that appears both in peak and wig files
		while(wig != null)
		{
			curwig=wig.nextChrom(curwig.getChr());
			if(curwig==null)
			{
				return null;
			}
			try{
				peak.moveToChr(curwig.getChr());
				System.out.println("Start processing " + curwig.getChr() + " peaks");
				break;
			   }
			catch (ItemNotFoundException e){
				System.out.println(e.getMessage());
			}
		}
		return curwig;
	}
		 
	public void splitPeaksWigDirectory(String peakFile, String wigFolder,float bd) throws Exception {
		m_bd = bd;
		boolean isZip=false;
		
		File dir = new File(wigFolder);
		if(!dir.exists()) {
			throw new FileNotFoundException("error occurred: wig folder " + wigFolder + " was not found");
		}
		
		File peakfile = new File(peakFile);
		RandomAccessFile buf = new RandomAccessFile(peakfile,"rw");
		PeakParser peak = new PeakParser(buf);
		
		File curwigfile = null;
		BufferedReader wigbuf = null;
		
		//loop over all peaks in file
		Position curPeak = peak.nextPosition();
		String curChr = "";
		String prevChr = "";
		Parser wig=null;
		
		long totalLength = peakfile.length();
		int times=0;
		System.out.println("Processing...");
		while(curPeak!=null){
			times++;
			if(times % 100 == 0)
        	{
				int progress = (int)(peak.getCurrentOffset() *100 / totalLength);
        		fireReadProgressEvent(progress);
        	}
			prevChr = curChr;
			curChr = curPeak.getChr();
			//read new wig file
			if(!curChr.equals(prevChr)){
				if(wigbuf!=null){
					System.out.println("End processing " + prevChr + " peaks");
				wigbuf.close();
				if(isZip){
					curwigfile.delete();
					isZip=false;
					}
				}
				String wigFileName = getRelatedChrFile(curChr,dir);
				   if(wigFileName.endsWith(".gz") || wigFileName.endsWith("zip")){
					  System.out.println( "Unzipping" + wigFileName);
					   String unzipFileName = this.Unzip(wigFileName,false);
					   wigFileName = unzipFileName;
					   isZip=true;
				   }
				   curwigfile = new File(wigFileName);
				   FileReader reader = new FileReader(wigFileName);
				   wigbuf = new BufferedReader(reader);
				   
				   
				   String wigType = Parser.getFileFormat(wigbuf);
					if(wigType.equals("wiggle_0"))
					{
						wig = new VariableStepParser(wigbuf,curChr);
					}
					else if(wigType.equals("bedGraph"))
					{
						 wig = new BedgraphParser(wigbuf);
					}
					else
					{
						if(wigType.equals(""))
						{
							throw new IOException ("error occurred: Can't guess the wig file's type, track type definition is missing\n");
						}
						else
						{
							throw new IOException("error occurred: unknown wig file type" + wigType+"\n");
						}
					}
			}
			splitPeak(curPeak,wig);
			curPeak = peak.nextPosition();
		}
		//close the last file
		System.out.println("End processing " + curChr + " peaks");
		wigbuf.close();
		if(isZip){
			curwigfile.delete();
		}
		fireReadProgressEvent(100);
	}
	private String getRelatedChrFile(String chrString, File dir) throws IOException {
		final String currentchr = chrString;
		
		FilenameFilter filter = new FilenameFilter() {
	        public boolean accept(File dir, String name) {
	        	return name.contains(currentchr);
	        }
	};
	String[] children = dir.list(filter);
	//sort the filename list in case more than one file contains the chromosome for example file.chr1 and file.chr11, chr1 will be the first
	java.util.Arrays.sort(children);
	   if (children.length<1) {
		   //Either directory does not exist or is not a directory
		   throw new FileNotFoundException ("error occurred: wig file for chromosome " + currentchr + " is missing\n" + "Make sure that the filename contains the string \"chr\" before the chromosome number");
		}
	   
	   String curFileName = dir.getAbsolutePath() + "/" + children[0];
	   System.out.println("Start processing " + chrString + " peaks");
	   System.out.println("Using wig file " + children[0]);
	   return curFileName;
	}
	
	
	private void splitPeak(Position peak, Parser wig) throws Exception{
		
		wig.m_bufRead.reset();
		ArrayList<wigPosition> wigPos = getOverlapList(peak,wig);
		//check if wigPos is not empty, in case there are not signal in wig file to specific peak
		if(wigPos.isEmpty())
		{
			System.out.println("no signal in wig file was found for peak "
					+ peak.getChr() + " " + peak.getStart() + " "
					+ peak.getEnd());
			System.out.println("continue to next peak");
			return;
		}
		
		wigPosition[] wigPosArr = new wigPosition[wigPos.size()];
		wigPosArr = wigPos.toArray(wigPosArr);
		
		boolean wayUp = true;
		boolean foundMaxima1 = false;
		boolean foundLocalValley = false;
		int localMaxima1=0;
		int localMaxima2=0;
		int localValley=0;
		int subPeakStart=0;
		
		if(wigPosArr.length == 1)
		{
			if(wigPosArr[0].getScore()>=m_cutoff)
			{
				SubPeak subPeak = new SubPeak(peak.getChr(),
						peak.getStart(),
						peak.getEnd(),
						wigPosArr[0].getScore(),
						peak.getStart(),
						peak.getEnd());
					subPeak.Print(m_fw);
					AddToSortedSet(subPeak);
			}
			return;
		}
		
		for(int i=0;i<wigPosArr.length;i++){
			if(i==0)
			{
				//first position
				localMaxima1 = 0;
				subPeakStart = 0;
				if(wigPosArr[1].getScore() > wigPosArr[0].getScore())
				{
					wayUp = true;
				}
				else
				{
					wayUp = false;
					foundMaxima1=true;
				}
				continue;
			}
			
			if(wayUp)
			{
				if(wigPosArr[i].getScore() >= wigPosArr[i-1].getScore())
				{
					continue;
				}
				else if(wigPosArr[i].getScore() < wigPosArr[i-1].getScore())
				{
					wayUp = false;
					if(!foundMaxima1)
					{
						localMaxima1 = i-1;
						foundMaxima1 = true;
					}
					else
					{
						localMaxima2 = i-1;
						float lowerSummitScore = 0f;
						//check who is higher
						if(wigPosArr[localMaxima1].getScore() >= wigPosArr[localMaxima2].getScore())
						{
							lowerSummitScore = wigPosArr[localMaxima2].getScore();
						}
						else
						{
							lowerSummitScore = wigPosArr[localMaxima1].getScore();
						}
						
						if(lowerSummitScore * m_bd > wigPosArr[localValley].getScore() )
						{
							//significant decrease split peaks
							int summitStart = wigPosArr[localMaxima1].getStart();
							int summitEnd = wigPosArr[localMaxima1].getEnd();
							int k=localMaxima1 - 1;
							while(k>0)
							{
								if(wigPosArr[k].getScore() == wigPosArr[localMaxima1].getScore()){
									summitStart = wigPosArr[k].getStart();
									k--;
								}
								else
									break;
							}
							if(wigPosArr[localMaxima1].getScore()>=m_cutoff)
							{
								SubPeak subPeak = new SubPeak(peak.getChr(),
										wigPosArr[subPeakStart].getStart(),
										wigPosArr[localValley].getStart(),
										wigPosArr[localMaxima1].getScore(),
										summitStart,
										summitEnd);
								subPeak.Print(m_fw);
								AddToSortedSet(subPeak);
							}
							for(int j=localValley+1; j< wigPosArr.length;j++)
							{
								if(wigPosArr[j].getScore()>0){
									subPeakStart = j;
									break;
								}
							}
							localMaxima1 = localMaxima2;
							foundLocalValley = false;
						}
						else //no split
						{ 
							if(wigPosArr[localMaxima2].getScore() >= wigPosArr[localMaxima1].getScore())
							{
								localMaxima1 = localMaxima2;
								foundLocalValley = false;
							}
							continue;
						}
					}
				}
			}
			else //way down
			{
				if(wigPosArr[i].getScore() < wigPosArr[i-1].getScore())
				{
					continue;
				}
				else if(wigPosArr[i].getScore() > wigPosArr[i-1].getScore())
				{	
					if(!foundLocalValley || wigPosArr[i-1].getScore() < wigPosArr[localValley].getScore())
					{
						//set only if local valley is lower than the current local valley
						localValley = i-1;
						foundLocalValley = true;
					}
					wayUp = true;
				}
			}
		}
		//handle last case
		if(wayUp)
		{
			if(!foundMaxima1 || wigPosArr[wigPosArr.length - 1].getScore() > wigPosArr[localMaxima1].getScore())
			{
				localMaxima1 = wigPosArr.length - 1;
			}
		}
		int summitStart = wigPosArr[localMaxima1].getStart();
		int summitEnd = wigPosArr[localMaxima1].getEnd();
		int k=localMaxima1 -1;
		while(k>0)
		{
			if(wigPosArr[k].getScore() == wigPosArr[localMaxima1].getScore()){
				summitStart = wigPosArr[k].getStart();
				k--;
			}
			else
				break;
		}
		if(wigPosArr[localMaxima1].getScore() >= m_cutoff)
		{
			SubPeak subPeak = new SubPeak(peak.getChr(),
					wigPosArr[subPeakStart].getStart(),
					wigPosArr[wigPosArr.length - 1].getEnd(),
					wigPosArr[localMaxima1].getScore(),
					summitStart,
					summitEnd);
			subPeak.Print(m_fw);
			AddToSortedSet(subPeak);
		}
	}

	private ArrayList<wigPosition> getOverlapList(Position peak, Parser wig) throws IOException {
		//get arraylist and fill holes with zero
		wigPosition prev = null;
		ArrayList<wigPosition> wigSignal = new ArrayList<wigPosition>();
		wigPosition p = (wigPosition) wig.nextPosition();
		while(p!=null && p.getChr().equals(peak.getChr())){
			
			if(p.contains(peak)==0){ //overlap
				//check if there is a hole
				if(prev!=null){
					if(p.getStart()!=prev.getEnd()+1 && p.getStart()!=prev.getEnd())
					{
							//there is a hole
							wigPosition hole = new wigPosition(p.getChr(),prev.getEnd()+1,p.getStart()-1,0);
							wigSignal.add(hole);
					}
				}
				//check is the first signal is larder than 0.0
				if(p.getScore()>0.0){
				wigSignal.add(p);
				}
				//update prev only when we reached to a region in wig that overlaps the peak
				prev=p;
				p = (wigPosition) wig.nextPosition();
			}
			else if (p.contains(peak)==1){
				//signal is after the peak
				wig.Lineback();
				break;
			}
			else{
				//signal before peak
				p = (wigPosition) wig.nextPosition();
				continue;
			}
		}
		
		return wigSignal;
	}
	
	
	private  String Unzip(String inFilePath,boolean throwEvent) throws IOException
	{
		File input = new File(inFilePath);
		String parentPath = input.getParent();
		
		String outFilePath=null;
		byte[] buf = new byte[102400];
	    int len;
	    OutputStream out=null;
	    
	    
	    if(inFilePath.endsWith(".gz"))
		{
			GZIPInputStream gzipInputStream = new GZIPInputStream(new FileInputStream(inFilePath),10000);
			outFilePath = inFilePath.replace(".gz", "");
			out = new FileOutputStream(outFilePath);
			long totalSize = input.length();
			int progress = 0;
			long readBytes = 0;
			 while ((len = gzipInputStream.read(buf,0,buf.length)) > 0)
			 {
				 	readBytes += 10000;
				 	progress = (int)(readBytes*100/totalSize);
			        out.write(buf, 0, len);
			        if(throwEvent)
			        	fireUnzipProgressEvent(progress);
			 }
			 gzipInputStream.close();
		}
		else
		{
			ZipFile zipFile = new ZipFile(inFilePath);
			Enumeration entries = zipFile.entries();
			
			if(zipFile.size()==0)
			{
				throw new IOException("error occurred: input zip file " + inFilePath + " is empty\n");
			}
			
			else if(zipFile.size()>1)
			{
				throw new IOException("error occurred: input zip file " + inFilePath + " contain more than one file\n");
			}

			while(entries.hasMoreElements()) 
			{
		        ZipEntry entry = (ZipEntry)entries.nextElement();

		        if(entry.isDirectory()) 
		        {
		          // Assume directories are stored parents first then children.
		         throw new IOException("error occurred: "+ entry.getName() + " is a directory, stop extracting zip file\n");
		        } 
		         outFilePath = entry.getName();
		        if(parentPath!=null)
		        {
		        	outFilePath = parentPath+File.separator+entry.getName();
		        }
		        InputStream in = zipFile.getInputStream(entry);
		        out = new FileOutputStream(outFilePath);
		        long totalSize = entry.getSize(); //uncompressed size
				int progress = 0;
				long readBytes = 0;
		        while ((len = in.read(buf)) > 0)
		        {
		        	readBytes += len;
				 	progress = (int)(readBytes*100/totalSize);
		        	out.write(buf, 0, len);   
		        	fireUnzipProgressEvent(progress);
		        }
		        in.close();
			}
		}
	    out.close();
	    return outFilePath;
	}
	
	public void clean() throws IOException {
		m_fw.close();
	}
	private void AddToSortedSet(SubPeak subPeak)
	{
		int position = Collections.binarySearch(m_SortedList,subPeak);
		if(position < 0)
		{
			position = (position+1) * -1;
		}
		
		m_SortedList.add(position,subPeak);
		if(m_SortedList.size() > m_numOfseq)
		{
			m_SortedList.remove(0);
		}
	}

	public void getPeakFasta(String dasURL, int seqLength) throws IOException, InterruptedException
	{
		System.out.println("Starting to fetch subpeaks sequences\n");
		String fastaOut = m_OutDir+ m_prefix +m_PeakFileNoExtension+".bestSubPeaks.fa";
		FileWriter fasta = new FileWriter(fastaOut);

		//go over the sorted peak array, and get the sequence for each binding site
		int count=0;
		Collections.reverse(m_SortedList);
		for(SubPeak p : m_SortedList)
		{
			count++;
			if(count%10==0)
			{
				System.out.println(count + " sequences have been fetched");
			}
			int summitPos = (p.m_summitStart+p.m_summitEnd)/2;
			int start = summitPos-(seqLength/2);
			int end = summitPos+(seqLength/2);
			String chr="";
			if(p.m_chr.toUpperCase().startsWith("CHR"))
			{
				chr=p.m_chr.toUpperCase().substring(3);
			}
			else
			{
				chr=p.m_chr.toUpperCase();
			}
			//change chrM to be chrMT
			if(chr.equals("M"))
			{
				chr = "MT";
			}
		
			String url = dasURL+ "/sequence?segment=" +chr +":"+start+","+end;
			URL ensemblDas = new URL(url);
		    URLConnection ensemblDasConnection = ensemblDas.openConnection();
		    try
		    {
		    	ensemblDasConnection.connect();
		    }
		    catch (UnknownHostException e) 
		    {
                System.out
						.println("error occurred: Can't connect to "+ dasURL +" to fetch sequences, check your internet connection");
                return;
		    }
		    catch (IOException e) 
		    {
		    	System.out
				.println("error occurred: Can't reach " + dasURL);
		    	return;
		    }


			BufferedReader br = new BufferedReader(new InputStreamReader(ensemblDasConnection.getInputStream()));
			String inputLine;
			String seq = "";
			boolean inSeq = false;
			while ((inputLine=br.readLine())!=null){
				if(inputLine.startsWith("</SEQUENCE>")){
					break;
				}
				if(inSeq){
					seq = seq+inputLine.trim();
				}
				else if(inputLine.startsWith("<SEQUENCE")){
					inSeq=true;
					seq = br.readLine();			
					}
				
			}
			if(seq.equals(""))
			{
				count--;
				System.out.println("Couldn't get sequence " + p.m_chr + ":"
						+ start + "-" + end + " from " + dasURL + "\n");
			}
			else
			{
				fasta.write(">" + p.m_chr+":"+start+"-"+end+"\n"+seq+"\n");
			}
		}
		System.out.println(count + " sequences have been fetched\n");
		fasta.close();	
	}
	public String getSubpeakFileName()
	{
		File temp = new File(m_SubpeakFileName);
		return temp.getName();
	}
	public String getFatsaName()
	{
		return m_prefix+m_PeakFileNoExtension+".bestSubPeaks.fa";
	}
	// This methods allows classes to register for unZipProgressEvent
    public void adduUnzipProgressEventListener(ProgressEventListener listener) {
        unzipListenerList.add(ProgressEventListener.class, listener);
    }

    // This methods allows classes to unregister for unZipProgressEvent
    public void removeUnzipProgressEventListener(ProgressEventListener listener) {
       unzipListenerList.remove(ProgressEventListener.class, listener);
    }

    // This private class is used to fire unZipProgressEvent
    void fireUnzipProgressEvent(int progress)
    {
    	if(progress > 100)
    		progress = 100;
    	ProgressEvent evt = new ProgressEvent(this, progress);
        Object[] listeners = unzipListenerList.getListenerList();
        // Each listener occupies two elements - the first is the listener class
        // and the second is the listener instance
        for (int i=0; i<listeners.length; i+=2) {
            if (listeners[i]==ProgressEventListener.class) {
                ((ProgressEventListener)listeners[i+1]).progressEventOccurred(evt);
            }
        }
    }
    
 // This methods allows classes to register for unZipProgressEvent
    public void addReadProgressEventListener(ProgressEventListener listener) {
        readListenerList.add(ProgressEventListener.class, listener);
    }

    // This methods allows classes to unregister for unZipProgressEvent
    public void removeReadProgressEventListener(ProgressEventListener listener) {
       readListenerList.remove(ProgressEventListener.class, listener);
    }

    // This private class is used to fire unZipProgressEvent
    void fireReadProgressEvent(int progress)
    {
    	if(progress > 100)
    		progress=100;
    	ProgressEvent evt = new ProgressEvent(this,progress);
        Object[] listeners = readListenerList.getListenerList();
        // Each listener occupies two elements - the first is the listener class
        // and the second is the listener instance
        for (int i=0; i<listeners.length; i+=2) {
            if (listeners[i]==ProgressEventListener.class) {
                ((ProgressEventListener)listeners[i+1]).progressEventOccurred(evt);
            }
        }
    }
    
}