package org.SplitPeaks;

import java.io.BufferedReader;
import java.io.IOException;

public class VariableStepParser extends Parser {

	private int m_space;
	private int m_sigDigit;
	private String m_chr;
	
	public VariableStepParser(BufferedReader reader, String chr) throws IOException{
		this(reader);
		if(m_chr==null){
			m_chr = chr;
		}	
		//check compatibility between chrM and chrMT, in case chr=MT, and m_chr=M
		if(chr.toLowerCase().equals("chrmt")&&m_chr.toLowerCase().equals("chrm"))
		{
			m_chr=chr;
		}
	}
	
	public VariableStepParser(BufferedReader reader) throws IOException {
    	super(reader,"WIG");
    	m_space = 1;
    	m_bufRead.mark(1000);
    	String line = m_bufRead.readLine();
    	if(line.contains("variableStep"))
    	{
    		m_space=1;
    		
    		String[] myelements = this.split(line);
    		for(String s : myelements)
    		{
    			if(s.toLowerCase().contains("chrom"))
    			{
    				try
    				{
    					m_chr = s.split("=")[1];
    				}
    				catch (Exception e)
    				{
    					throw new IOException("Invalid variableStep header line \"" + line + "\"");
    				}
    			}
    			else if(s.toLowerCase().contains("span"))
				{
    				try
    				{
    					m_space = Integer.parseInt(s.split("=")[1]);
    				}
    				catch (Exception e)
    				{
    					throw new IOException("Invalid variableStep header line \"" + line + "\"");
    				}
				}

    		}
    		
    		m_bufRead.mark(1000);
    	}
    	line = m_bufRead.readLine();
		String[] myelements = this.split(line);
        if(myelements.length !=2) {
        	throw new IOException ("Invalid variableStep wig file format");
        }
        
        int signal =0;
        try
        {
        	signal = Integer.parseInt(myelements[0]);
        	m_sigDigit=signal%10;
        }
        catch (Exception e)
        {
        	throw new IOException ("Invalid variableStep wig file format in line starting with \"" + line + "\"" + "\n");
        }
        
        //return to be before the first line of signal
        m_bufRead.reset();
	}
	
    public  wigPosition nextPosition () throws IOException 
    {
    	m_bufRead.mark(1000);
		String line = m_bufRead.readLine();
		if(line==null || line.equals(""))
		{
			return null;
		}
		m_currentOffset += line.length();
		wigPosition pos=null;
			String[] myelements = this.split(line);
			if(line.contains("chrom"))
			{
				m_space=1;
				
	    		for(String s : myelements)
	    		{
	    			if(s.toLowerCase().contains("chrom"))
	    			{
	    				try
	    				{
	    					m_chr = s.split("=")[1];
	    				}
	    				catch (Exception e)
	    				{
	    					throw new IOException("Invalid variableStep header line \"" + line + "\"");
	    				}
	    			}
	    			
	    			else if(s.toLowerCase().contains("span"))
    				{
	    				try
	    				{
	    					m_space = Integer.parseInt(s.split("=")[1]);
	    				}
	    				catch (Exception e)
	    				{
	    					throw new IOException("Invalid variableStep header line \"" + line + "\"");
	    				}
    				}	
	    		}
				line = m_bufRead.readLine();
				myelements = this.split(line);
			}
			int start=0;
			int end=0;
			float score = 0;
			
			try
			{
				start = Integer.parseInt(myelements[0]);
				end = start+m_space-1;
				score = Float.parseFloat(myelements[1]);
			}
			catch (Exception e)
			{
    			throw new IOException ("Invalid variableStep wig file format in line starting with \"" + line + "\"" + "\n");
			}		
			pos = new wigPosition(m_chr,start,end,score);
			return pos;
			
			                
	}
    public wigPosition nextChrom(String curChr) throws IOException{
		
		wigPosition pos = nextPosition();
		while(pos != null && pos.getChr().equals(curChr))
		{
			pos = nextPosition();
		}
		return pos;
	}
    
	public int getDigit() {
		return m_sigDigit;
	}
}
